@@ -2,7 +2,7 @@
  *
  * pqxx::connection encapsulates a connection to a database.
  *
- * Copyright (c) 2000-2021, Jeroen T. Vermeulen.
+ * Copyright (c) 2000-2022, Jeroen T. Vermeulen.
  *
  * See COPYING for copyright license.  If you did not receive a file called
  * COPYING with this source code, please notify the distributor of this
@@ -12,7 +12,6 @@
 
 #include <algorithm>
 #include <array>
-#include <cassert>
 #include <cerrno>
 #include <cstdio>
 #include <cstdlib>
@@ -22,58 +21,43 @@
 #include <iterator>
 #include <memory>
 #include <stdexcept>
+#include <utility>
 
-// For WSAPoll():
-#if __has_include(<winsock2.h>)
-#  include <winsock2.h>
-#endif
-#if __has_include(<ws2tcpip.h>)
-#  include <ws2tcpip.h>
-#endif
-#if __has_include(<mstcpip.h>)
-#  include <mstcpip.h>
-#endif
-
-// For poll():
-#if __has_include(<poll.h>)
-#  include <poll.h>
-#endif
-
-// For select() on recent POSIX systems.
-#if __has_include(<sys/select.h>)
-#  include <sys/select.h>
-#endif
-
-// For select() on some older POSIX systems.
-#if __has_include(<sys/types.h>)
-#  include <sys/types.h>
+// For fcntl().
+#if __has_include(<fcntl.h>)
+#  include <fcntl.h>
 #endif
 #if __has_include(<unistd.h>)
 #  include <unistd.h>
 #endif
-#if __has_include(<sys/time.h>)
-#  include <sys/time.h>
-#endif
 
+// For ioctlsocket().
+#if defined(_WIN32) && __has_include(<winsock2.h>)
+#  include <winsock2.h>
+#endif
 
 extern "C"
 {
 #include <libpq-fe.h>
 }
 
-#include "pqxx/binarystring"
-#include "pqxx/config-internal-libpq.h"
-#include "pqxx/nontransaction"
-#include "pqxx/notification"
-#include "pqxx/pipeline"
-#include "pqxx/result"
-#include "pqxx/strconv"
-#include "pqxx/transaction"
+#include "pqxx/internal/header-pre.hxx"
+
+#include "pqxx/binarystring.hxx"
+#include "pqxx/internal/wait.hxx"
+#include "pqxx/nontransaction.hxx"
+#include "pqxx/notification.hxx"
+#include "pqxx/pipeline.hxx"
+#include "pqxx/result.hxx"
+#include "pqxx/strconv.hxx"
+#include "pqxx/transaction.hxx"
 
 #include "pqxx/internal/gates/errorhandler-connection.hxx"
 #include "pqxx/internal/gates/result-connection.hxx"
 #include "pqxx/internal/gates/result-creation.hxx"
 
+#include "pqxx/internal/header-post.hxx"
+
 
 extern "C"
 {
@@ -92,11 +76,12 @@ extern "C"
 
 using namespace std::literals;
 
-std::string pqxx::encrypt_password(char const user[], char const password[])
+std::string PQXX_COLD
+pqxx::encrypt_password(char const user[], char const password[])
 {
   std::unique_ptr<char, std::function<void(char *)>> p{
     PQencryptPassword(password, user), PQfreemem};
-  return std::string{p.get()};
+  return {p.get()};
 }
 
 
@@ -108,13 +93,45 @@ pqxx::connection::connection(connection &&rhs) :
 }
 
 
+pqxx::connection::connection(
+  connection::connect_mode, zview connection_string) :
+        m_conn{PQconnectStart(connection_string.c_str())}
+{
+  if (m_conn == nullptr)
+    throw std::bad_alloc{};
+  if (status() == CONNECTION_BAD)
+    throw pqxx::broken_connection{PQerrorMessage(m_conn)};
+}
+
+
+std::pair<bool, bool> pqxx::connection::poll_connect()
+{
+  switch (PQconnectPoll(m_conn))
+  {
+  case PGRES_POLLING_FAILED:
+    throw pqxx::broken_connection{PQerrorMessage(m_conn)};
+  case PGRES_POLLING_READING: return std::make_pair(true, false);
+  case PGRES_POLLING_WRITING: return std::make_pair(false, true);
+  case PGRES_POLLING_OK:
+    if (not is_open())
+      throw pqxx::broken_connection{PQerrorMessage(m_conn)};
+    return std::make_pair(false, false);
+  case PGRES_POLLING_ACTIVE:
+    throw internal_error{
+      "Nonblocking connection poll returned obsolete 'active' state."};
+  default:
+    throw internal_error{
+      "Nonblocking connection poll returned unknown value."};
+  }
+}
+
 void pqxx::connection::complete_init()
 {
   if (m_conn == nullptr)
     throw std::bad_alloc{};
   try
   {
-    if (PQstatus(m_conn) != CONNECTION_OK)
+    if (not is_open())
       throw broken_connection{PQerrorMessage(m_conn)};
 
     set_up_state();
@@ -122,6 +139,7 @@ void pqxx::connection::complete_init()
   catch (std::exception const &)
   {
     PQfinish(m_conn);
+    m_conn = nullptr;
     throw;
   }
 }
@@ -176,11 +194,9 @@ pqxx::connection &pqxx::connection::operator=(connection &&rhs)
 
   close();
 
-  m_conn = rhs.m_conn;
+  m_conn = std::exchange(rhs.m_conn, nullptr);
   m_unique_id = rhs.m_unique_id;
 
-  rhs.m_conn = nullptr;
-
   return *this;
 }
 
@@ -196,14 +212,25 @@ pqxx::result pqxx::connection::make_result(
     else
       throw broken_connection{"Lost connection to the database server."};
   }
-  auto const r{pqxx::internal::gate::result_creation::create(
-    pgr, query, internal::enc_group(encoding_id()))};
+  internal::encoding_group enc;
+  try
+  {
+    enc = internal::enc_group(encoding_id());
+  }
+  catch (std::exception const &)
+  {
+    // Don't let the PGresult leak.
+    // TODO: Can we just accept a unique_ptr instead?
+    internal::clear_result(pgr);
+    throw;
+  }
+  auto const r{pqxx::internal::gate::result_creation::create(pgr, query, enc)};
   pqxx::internal::gate::result_creation{r}.check_status(desc);
   return r;
 }
 
 
-int pqxx::connection::backendpid() const noexcept
+int PQXX_COLD pqxx::connection::backendpid() const &noexcept
 {
   return (m_conn == nullptr) ? 0 : PQbackendPID(m_conn);
 }
@@ -218,34 +245,45 @@ PQXX_PURE int socket_of(::pqxx::internal::pq::PGconn const *c) noexcept
 } // namespace
 
 
-int pqxx::connection::sock() const noexcept
+int pqxx::connection::sock() const &noexcept
 {
   return socket_of(m_conn);
 }
 
 
-int pqxx::connection::protocol_version() const noexcept
+int PQXX_COLD pqxx::connection::protocol_version() const noexcept
 {
   return (m_conn == nullptr) ? 0 : PQprotocolVersion(m_conn);
 }
 
 
-int pqxx::connection::server_version() const noexcept
+int PQXX_COLD pqxx::connection::server_version() const noexcept
 {
   return PQserverVersion(m_conn);
 }
 
 
 void pqxx::connection::set_variable(
-  std::string_view var, std::string_view value)
+  std::string_view var, std::string_view value) &
 {
-  exec(internal::concat("SET ", var, "=", value));
+  exec(internal::concat("SET ", quote_name(var), "=", value));
 }
 
 
 std::string pqxx::connection::get_variable(std::string_view var)
 {
-  return exec(internal::concat("SHOW ", var)).at(0).at(0).as(std::string{});
+  return exec(internal::concat("SHOW ", quote_name(var)))
+    .at(0)
+    .at(0)
+    .as(std::string{});
+}
+
+
+std::string pqxx::connection::get_var(std::string_view var)
+{
+  // (Variables can't be null, so far as I can make out.)
+  return exec(internal::concat("SHOW "sv, quote_name(var)))[0][0]
+    .as<std::string>();
 }
 
 
@@ -338,7 +376,7 @@ void pqxx::connection::process_notice(zview msg) noexcept
 }
 
 
-void pqxx::connection::trace(FILE *out) noexcept
+void PQXX_COLD pqxx::connection::trace(FILE *out) noexcept
 {
   if (m_conn)
   {
@@ -350,7 +388,7 @@ void pqxx::connection::trace(FILE *out) noexcept
 }
 
 
-void pqxx::connection::add_receiver(pqxx::notification_receiver *n)
+void PQXX_COLD pqxx::connection::add_receiver(pqxx::notification_receiver *n)
 {
   if (n == nullptr)
     throw argument_error{"Null receiver registered"};
@@ -374,7 +412,8 @@ void pqxx::connection::add_receiver(pqxx::notification_receiver *n)
 }
 
 
-void pqxx::connection::remove_receiver(pqxx::notification_receiver *T) noexcept
+void PQXX_COLD
+pqxx::connection::remove_receiver(pqxx::notification_receiver *T) noexcept
 {
   if (T == nullptr)
     return;
@@ -422,7 +461,7 @@ bool pqxx::connection::is_busy() const noexcept
 }
 
 
-void pqxx::connection::cancel_query()
+void PQXX_COLD pqxx::connection::cancel_query()
 {
   using pointer = std::unique_ptr<PGcancel, std::function<void(PGcancel *)>>;
   pointer cancel{PQgetCancel(m_conn), PQfreeCancel};
@@ -440,7 +479,88 @@ void pqxx::connection::cancel_query()
 }
 
 
-void pqxx::connection::set_verbosity(error_verbosity verbosity) noexcept
+namespace
+{
+// C++20: std::span?
+/// Get error string for a given @c errno value.
+template<std::size_t BYTES>
+char const *PQXX_COLD
+error_string(int err_num, std::array<char, BYTES> &buffer)
+{
+  // Not entirely clear whether strerror_s will be in std or global namespace.
+  using namespace std;
+
+#if defined(PQXX_HAVE_STERROR_S) || defined(PQXX_HAVE_STRERROR_R)
+#  if defined(PQXX_HAVE_STRERROR_S)
+  auto const err_result{strerror_s(std::data(buffer), BYTES, err_num)};
+#  else
+  auto const err_result{strerror_r(err_num, std::data(buffer), BYTES)};
+#  endif
+  if constexpr (std::is_same_v<pqxx::strip_t<decltype(err_result)>, char *>)
+  {
+    // GNU version of strerror_r; returns the error string, which may or may
+    // not reside within buffer.
+    return err_result;
+  }
+  else
+  {
+    // Either strerror_s or POSIX strerror_r; returns an error code.
+    // Sorry for being lazy here: Not reporting error string for the case
+    // where we can't retrieve an error string.
+    if (err_result == 0)
+      return std::data(buffer);
+    else
+      return "Compound errors.";
+  }
+
+#else
+  // Fallback case, hopefully for no actual platforms out there.
+  pqxx::ignore_unused(err_num, buffer);
+  return "(No error information available.)";
+#endif
+}
+} // namespace
+
+
+#if defined(_WIN32) || __has_include(<fcntl.h>)
+void pqxx::connection::set_blocking(bool block) &
+{
+  auto const fd{sock()};
+#  if defined _WIN32
+  unsigned long mode{not block};
+  if (::ioctlsocket(fd, FIONBIO, &mode) != 0)
+  {
+    std::array<char, 200> errbuf;
+    char const *err{error_string(WSAGetLastError(), errbuf)};
+    throw broken_connection{
+      internal::concat("Could not set socket's blocking mode: ", err)};
+  }
+#  else  // _WIN32
+  std::array<char, 200> errbuf;
+  auto flags{::fcntl(fd, F_GETFL, 0)};
+  if (flags == -1)
+  {
+    char const *const err{error_string(errno, errbuf)};
+    throw broken_connection{
+      internal::concat("Could not get socket state: ", err)};
+  }
+  if (block)
+    flags |= O_NONBLOCK;
+  else
+    flags &= ~O_NONBLOCK;
+  if (::fcntl(fd, F_SETFL, flags) == -1)
+  {
+    char const *const err{error_string(errno, errbuf)};
+    throw broken_connection{
+      internal::concat("Could not set socket's blocking mode: ", err)};
+  }
+#  endif // _WIN32
+}
+#endif // defined(_WIN32) || __has_include(<fcntl.h>)
+
+
+void PQXX_COLD
+pqxx::connection::set_verbosity(error_verbosity verbosity) &noexcept
 {
   PQsetErrorVerbosity(m_conn, static_cast<PGVerbosity>(verbosity));
 }
@@ -463,8 +583,7 @@ notify_ptr get_notif(pqxx::internal::pq::PGconn *conn)
 int pqxx::connection::get_notifs()
 {
   if (not consume_input())
-    PQXX_UNLIKELY
-  throw broken_connection{"Connection lost."};
+    throw broken_connection{"Connection lost."};
 
   // Even if somehow we receive notifications during our transaction, don't
   // deliver them.
@@ -515,25 +634,25 @@ int pqxx::connection::get_notifs()
 }
 
 
-char const *pqxx::connection::dbname() const
+char const *PQXX_COLD pqxx::connection::dbname() const
 {
   return PQdb(m_conn);
 }
 
 
-char const *pqxx::connection::username() const
+char const *PQXX_COLD pqxx::connection::username() const
 {
   return PQuser(m_conn);
 }
 
 
-char const *pqxx::connection::hostname() const
+char const *PQXX_COLD pqxx::connection::hostname() const
 {
   return PQhost(m_conn);
 }
 
 
-char const *pqxx::connection::port() const
+char const *PQXX_COLD pqxx::connection::port() const
 {
   return PQport(m_conn);
 }
@@ -546,7 +665,7 @@ char const *pqxx::connection::err_msg() const noexcept
 }
 
 
-void pqxx::connection::register_errorhandler(errorhandler *handler)
+void PQXX_COLD pqxx::connection::register_errorhandler(errorhandler *handler)
 {
   // Set notice processor on demand, i.e. only when the caller actually
   // registers an error handler.
@@ -563,7 +682,8 @@ void pqxx::connection::register_errorhandler(errorhandler *handler)
 }
 
 
-void pqxx::connection::unregister_errorhandler(errorhandler *handler) noexcept
+void PQXX_COLD
+pqxx::connection::unregister_errorhandler(errorhandler *handler) noexcept
 {
   // The errorhandler itself will take care of nulling its pointer to this
   // connection.
@@ -573,10 +693,10 @@ void pqxx::connection::unregister_errorhandler(errorhandler *handler) noexcept
 }
 
 
-std::vector<pqxx::errorhandler *> pqxx::connection::get_errorhandlers() const
+std::vector<pqxx::errorhandler *>
+  PQXX_COLD pqxx::connection::get_errorhandlers() const
 {
-  return std::vector<errorhandler *>{
-    std::begin(m_errorhandlers), std::end(m_errorhandlers)};
+  return {std::begin(m_errorhandlers), std::end(m_errorhandlers)};
 }
 
 
@@ -622,10 +742,9 @@ std::string pqxx::connection::encrypt_password(
 }
 
 
-void pqxx::connection::prepare(char const name[], char const definition[])
+void pqxx::connection::prepare(char const name[], char const definition[]) &
 {
-  // Allocate once, re-use across invocations.
-  static auto const q{std::make_shared<std::string>(
+  auto const q{std::make_shared<std::string>(
     pqxx::internal::concat("[PREPARE ", name, "]"))};
 
   auto const r{
@@ -633,7 +752,7 @@ void pqxx::connection::prepare(char const name[], char const definition[])
 }
 
 
-void pqxx::connection::prepare(char const definition[])
+void pqxx::connection::prepare(char const definition[]) &
 {
   this->prepare("", definition);
 }
@@ -740,8 +859,6 @@ void pqxx::connection::unregister_transaction(transaction_base *t) noexcept
 std::pair<std::unique_ptr<char, std::function<void(char *)>>, std::size_t>
 pqxx::connection::read_copy_line()
 {
-  using raw_line =
-    std::pair<std::unique_ptr<char, std::function<void(char *)>>, std::size_t>;
   char *buf{nullptr};
 
   // Allocate once, re-use across invocations.
@@ -756,7 +873,7 @@ pqxx::connection::read_copy_line()
 
   case -1: // End of COPY.
     make_result(PQgetResult(m_conn), q, *q);
-    return raw_line{};
+    return {};
 
   case 0: // "Come back later."
     throw internal_error{"table read inexplicably went asynchronous"};
@@ -843,7 +960,7 @@ std::string pqxx::connection::esc(std::string_view text) const
 }
 
 
-std::string
+std::string PQXX_COLD
 pqxx::connection::esc_raw(unsigned char const bin[], std::size_t len) const
 {
   return pqxx::internal::esc_bin(binary_cast(bin, len));
@@ -857,7 +974,7 @@ pqxx::connection::esc_raw(std::basic_string_view<std::byte> bin) const
 }
 
 
-std::string pqxx::connection::unesc_raw(char const text[]) const
+std::string PQXX_COLD pqxx::connection::unesc_raw(char const text[]) const
 {
   if (text[0] == '\\' and text[1] == 'x')
   {
@@ -882,7 +999,7 @@ std::string pqxx::connection::unesc_raw(char const text[]) const
 }
 
 
-std::string
+std::string PQXX_COLD
 pqxx::connection::quote_raw(unsigned char const bin[], std::size_t len) const
 {
   return internal::concat("'", esc_raw(binary_cast(bin, len)), "'::bytea");
@@ -896,7 +1013,7 @@ pqxx::connection::quote_raw(std::basic_string_view<std::byte> bytes) const
 }
 
 
-std::string pqxx::connection::quote(binarystring const &b) const
+std::string PQXX_COLD pqxx::connection::quote(binarystring const &b) const
 {
   return quote(b.bytes_view());
 }
@@ -954,104 +1071,13 @@ pqxx::connection::esc_like(std::string_view text, char escape_char) const
 }
 
 
-namespace
-{
-// Convert a timeval to milliseconds, or -1 if no timeval is given.
-[[maybe_unused]] constexpr int tv_milliseconds(timeval *tv = nullptr)
-{
-  if (tv == nullptr)
-    return -1;
-  else
-    return pqxx::check_cast<int>(
-      tv->tv_sec * 1000 + tv->tv_usec / 1000, "milliseconds"sv);
-}
-
-
-/// Wait for an fd to become free for reading/writing.  Optional timeout.
-void wait_fd(int fd, bool forwrite = false, timeval *tv = nullptr)
-{
-  if (fd < 0)
-    PQXX_UNLIKELY
-  throw pqxx::broken_connection{"No connection."};
-
-// WSAPoll is available in winsock2.h only for versions of Windows >= 0x0600
-#if defined(_WIN32) && (_WIN32_WINNT >= 0x0600)
-  short const events{forwrite ? POLLWRNORM : POLLRDNORM};
-  WSAPOLLFD fdarray{SOCKET(fd), events, 0};
-  WSAPoll(&fdarray, 1, tv_milliseconds(tv));
-  // TODO: Check for errors.
-#elif defined(PQXX_HAVE_POLL)
-  auto const events{static_cast<short>(
-    POLLERR | POLLHUP | POLLNVAL | (forwrite ? POLLOUT : POLLIN))};
-  pollfd pfd{fd, events, 0};
-  poll(&pfd, 1, tv_milliseconds(tv));
-  // TODO: Check for errors.
-#else
-  // No poll()?  Our last option is select().
-  fd_set read_fds;
-  FD_ZERO(&read_fds);
-  if (not forwrite)
-    FD_SET(fd, &read_fds);
-
-  fd_set write_fds;
-  FD_ZERO(&write_fds);
-  if (forwrite)
-    FD_SET(fd, &write_fds);
-
-  fd_set except_fds;
-  FD_ZERO(&except_fds);
-  FD_SET(fd, &except_fds);
-
-  select(fd + 1, &read_fds, &write_fds, &except_fds, tv);
-  // TODO: Check for errors.
-#endif
-}
-} // namespace
-
-void pqxx::internal::wait_read(internal::pq::PGconn const *c)
-{
-  wait_fd(socket_of(c));
-}
-
-
-void pqxx::internal::wait_read(
-  internal::pq::PGconn const *c, std::time_t seconds, long microseconds)
-{
-  // Not all platforms have suseconds_t for the microseconds.  And some 64-bit
-  // systems use 32-bit integers here.
-  timeval tv{
-    check_cast<decltype(timeval::tv_sec)>(seconds, "read timeout seconds"sv),
-    check_cast<decltype(timeval::tv_usec)>(
-      microseconds, "read timeout microseconds"sv)};
-  wait_fd(socket_of(c), false, &tv);
-}
-
-
-void pqxx::internal::wait_write(internal::pq::PGconn const *c)
-{
-  wait_fd(socket_of(c), true);
-}
-
-
-void pqxx::connection::wait_read() const
-{
-  internal::wait_read(m_conn);
-}
-
-
-void pqxx::connection::wait_read(std::time_t seconds, long microseconds) const
-{
-  internal::wait_read(m_conn, seconds, microseconds);
-}
-
-
 int pqxx::connection::await_notification()
 {
   int notifs = get_notifs();
   if (notifs == 0)
   {
     PQXX_LIKELY
-    wait_read();
+    internal::wait_fd(socket_of(m_conn), true, false, 10, 0);
     notifs = get_notifs();
   }
   return notifs;
@@ -1065,7 +1091,10 @@ int pqxx::connection::await_notification(
   if (notifs == 0)
   {
     PQXX_LIKELY
-    wait_read(seconds, microseconds);
+    internal::wait_fd(
+      socket_of(m_conn), true, false,
+      check_cast<unsigned>(seconds, "Seconds out of range."),
+      check_cast<unsigned>(microseconds, "Microseconds out of range."));
     return get_notifs();
   }
   return notifs;
@@ -1088,7 +1117,7 @@ std::string pqxx::connection::get_client_encoding() const
 }
 
 
-void pqxx::connection::set_client_encoding(char const encoding[])
+void PQXX_COLD pqxx::connection::set_client_encoding(char const encoding[]) &
 {
   switch (auto const retval{PQsetClientEncoding(m_conn, encoding)}; retval)
   {
@@ -1160,7 +1189,14 @@ char const *get_default(PQconninfoOption const &opt) noexcept
   }
   // As of C++11, std::getenv() uses thread-local storage, so it should be
   // thread-safe.  MSVC still warns about it though.
+#if defined(_MSC_VER)
+#  pragma warning(push)
+#  pragma warning(disable : 4996)
+#endif
   char const *var{std::getenv(opt.envvar)};
+#if defined(_MSC_VER)
+#  pragma warning(pop)
+#endif
   if (var == nullptr)
   {
     // There's an environment variable for this setting, but it's not set.
@@ -1206,3 +1242,32 @@ std::string pqxx::connection::connection_string() const
   }
   return buf;
 }
+
+
+#if defined(_WIN32) || __has_include(<fcntl.h>)
+pqxx::connecting::connecting(zview connection_string) :
+        m_conn{connection::connect_nonblocking, connection_string}
+{}
+#endif // defined(_WIN32) || __has_include(<fcntl.h>
+
+
+#if defined(_WIN32) || __has_include(<fcntl.h>)
+void pqxx::connecting::process() &
+{
+  auto const [reading, writing]{m_conn.poll_connect()};
+  m_reading = reading;
+  m_writing = writing;
+}
+#endif // defined(_WIN32) || __has_include(<fcntl.h>
+
+
+#if defined(_WIN32) || __has_include(<fcntl.h>)
+pqxx::connection pqxx::connecting::produce() &&
+{
+  if (!done())
+    throw usage_error{
+      "Tried to produce a nonblocking connection before it was done."};
+  m_conn.complete_init();
+  return std::move(m_conn);
+}
+#endif // defined(_WIN32) || __has_include(<fcntl.h>
